/* 
    Persona 3/4 Model Importer, version 4.1
    Written by TGE. Credit would be nice ;)
    Apologies if the script is a bit messy here and there, the format is kinda confusing :) 

    To use automatic texture importing & exporting you need to have RMDTexTool.exe in the same folder as the script!
    The program will dump the textures to the %filename%_tex\ folder in the same folder as the imported file!
    
    Changelog version 4.0:
        - Proper parsing of HAnim nodes, in order to help support the animation importer
        - User properties are set for bone nodes, containing: 
            - Name id
            - Frame index
            - HAnim node index
        - Bones are now given names based on the HAnim name id
        - Various performance improvements
        
    Changelog version 4.1
        - Implemented more compatible texture exporter
*/

global gUVFlip = true -- set to false if you want the original uv map
global gImportWeights = true -- set to false if you dont want to import the weighting (speeds up importing a lot)
global gImportNormals = true -- set to false if you dont want to import the normals (speeds up importing)

fn ReadFixedString bstream fixedLen =
(
   local str = ""
   for i = 1 to fixedLen do
   (
	  str0 = ReadByte bstream #unsigned
	  str+= bit.intAsChar str0
   )
   str
)

fn PrintChunk = ()

fn ReadChunk fstream =
(

	struct ChunkStruct
	(
		ChunkStart,
		DataType,
		DataSize,
		DataSeperator
	)

	Chunk = (ChunkStruct \
		ChunkStart:(ftell fstream) \
		DataType:(readlong fstream #unsigned) \
		DataSize:(readlong fstream #unsigned) \
		DataSeperator:(readlong fstream #unsigned) \ -- I presume this is a command
	)
    
    PrintChunk( Chunk )
	
	return Chunk
)

fn PrintChunk Chunk = 
(
	print ("Chunk 0x"+(bit.intashex(Chunk.DataType) as string)+" size 0x"+(bit.intashex(Chunk.DataSize) as string)+" @ 0x"+(bit.intashex(Chunk.ChunkStart) as string))
)

fn CalculateChunkEndOffset Chunk =
(
	ChunkEnd = ((Chunk.ChunkStart + 0xC) + Chunk.DataSize)
	return ChunkEnd
)

fn ReadMatProperty Chunk fstream =
(
	MatProperty_Array = #()
	while (not ftell fstream >= CalculateChunkEndOffset Chunk) do
	(
		prop = readlong fstream #unsigned
		if (prop != 0xFFFF) then append MatProperty_Array prop
	)
	return MatProperty_Array
)

struct ModelInfoStruct 
(
	ModelCount,
	Unknown1,
	Unknown2
)

struct MeshInfoStruct
(
	MeshCount
)

struct BoundingBoxStruct
(
	minX,
	minY,
	minZ,
	maxX,
	maxY,
	maxZ
)

struct UVStruct
(
	UV_Array
)

struct RootBoneStruct
(
	RootBoneArray
)

struct BoneStruct 
(
	BoneParent,
	BoneName,
	BoneTFM
)

struct SkinDataStruct
(
	RootBoneID,
	ModelID,
	Flag1,
	Flag2
)

struct WeightBoneIDs
(
	bone1,
	bone2,
	bone3,
	bone4
)

struct weight_data
(
	boneids,weights
)

struct WeightDataInfo
(
	UsedBoneCount,
	UsedBoneIndexCount,
	Weight_Array
)

struct MeshData
(
	Color_Array,
	UVs_Array,
	Face_Array,
	MatID_Array,
	Vert_Array,
	Normal_Array
)

struct HAnimNodeInfo
(
    NameId,
    Index,
    FrameIndex
)

clearlistener()
global fname = getOpenFileName \ 
caption:"Open Model" \
types:"Persona 3/4 Model (*.*)|*.*|*.*|*.*" \
historyCategory:"AtlusRMD Object Presets"
global st = timestamp() --get start time in milliseconds

global f = fopen fname "rb"
global filename = GetFilenameFile fname
global filepath = getFileNamePath fname
global filetype = getFilenameType fname
global scriptPath = getThisScriptFilename()
global scriptPath = getFileNamePath scriptPath

fseek f 0 #seek_end
global FileEnd = ftell f
fseek f 0 #seek_set
--print (("File Size = ")+(bit.intashex(fileEnd) as string))

global ModelInfoArray = #()
global MaterialGroupInfoArray = #()
global MaterialNameArray = #()
global MaterialNamesArray = #()
global ModelCounter = 0
global MeshCounter = 0
global RootBonesArray = #()
global BoneArray = #()
global BoneDataArray = #()
global WeightDataInfoArray = #()
global MeshDataArray = #()
global TexNameArray = #()
global BoneExtensionArray = #()
global BoneIndexStart = 0
global ModelIndexStart = 0
global MeshCount = 0
global BoneCount = 0
global MaterialCounter = 0
global hAnimNodeInfoArray = #()

case (fileType) of -- some actions based on the extension of the file, default is for RMDs
(
	(".PAC"): (
		fseek f 0x100 #seek_cur
		fseek f 0xC #seek_cur
		animCount = readshort f
	)
	(".FPC"): (
		fseek f 0x100 #seek_cur
	)
	(".P01"): (
		fseek f 0x100 #seek_cur
	)
	(".RWS"): (
	)
	default: (
		type = readlong f
		if type == 0xFFFFFFFFF0F000F0 then
		(
			fseek f 0x8 #seek_cur -- RMD header
			animCount = readshort f
		) else (fseek f 0 #seek_set)
	)
)

DisableSceneRedraw()

while (not ftell f >= fileEnd) do 
(
	SubChunkMain = ReadChunk f
	
	--PrintChunk SubChunkMain
	
	case (SubChunkMain.DataType) of 
	(
		(0x16): ( -- Texture Group
			while (not ftell f >= CalculateChunkEndOffset SubChunkMain) do
			(
				SubChunkTextureGroup = ReadChunk f
				--PrintChunk SubChunkTextureGroup
				texNameRead = 0
				
				case (SubChunkTextureGroup.DataType) of
				(
					(0x15): ( -- Texture Node
						while (not ftell f >= CalculateChunkEndOffset SubChunkTextureGroup) do
						(
							SubChunkTextureNode = ReadChunk f
							--PrintChunk SubChunkTextureNode
							case (SubChunkTextureNode.DataType) of
							(
								(0x2): ( -- Texture Name Info
									texNameRead += 1
									texName = readfixedstring f SubChunkTextureNode.DataSize
									if texNameRead == 1 then (
									append TexNameArray texName
									)
								)
								(0x1): ( -- Texture Image
									if (SubChunkTextureNode.DataSize >= 80) then -- 80 is the size of the texture header
									(
										texBuffer = #()
										totSize = SubChunkTextureGroup.DataSize + 0xC
										texBuffer[totSize] = 0
										fseek f SubChunkTextureGroup.ChunkStart #seek_set
										for i = 1 to (totSize) do texBuffer[i] = readbyte f
											
										png_out = filepath + filename + "_tex\\" + TexNameArray[TexNameArray.count] + ".png"
										tex_out = filepath + filename + "_tex\\" + TexNameArray[TexNameArray.count] + ".txn"
										makedir (filepath + filename + "_tex\\")

										if doesFileExist (png_out) == true then (
											format "PNG found! Skipping conversion\n\n"
										) -- if doesFileExist (png_out) == true then
										else
										(
											if doesFileExist (scriptPath + "TextureExporter\\RMDTextureExporter.exe") == true then
											(
												texOut = fopen tex_out "wb"
												for i = 1 to (totSize) do writebyte texOut texBuffer[i]
												fclose texOut
												ShellLaunch (scriptPath + "TextureExporter\\RMDTextureExporter.exe") ("\"" + filepath + filename + "_tex\\" + TexNameArray[TexNameArray.count] + ".txn" + "\"")
												sleep 0.2
												--deleteFile tex_out 
											) -- if doesFileExist (scriptPath + "RMDTexTool.exe") == true then
											else
											(
												format "RMDTexTool.exe not found! Skipping conversion\n\n"
											) -- if doesFileExist (scriptPath + "RMDTexTool.exe") == false then
										) -- if doesFileExist (png_out) == false then
                                        
                                        fseek f ( CalculateChunkEndOffset SubChunkTextureNode ) #seek_set
                                        
									) -- if (SubChunkTextureNode.DataSize >= 80) then
									else
									(
										fseek f SubChunkTextureNode.DataSize #seek_cur
									)
								)
								default: (
									fseek f SubChunkTextureNode.DataSize #seek_cur
								)
							)
						)
					)
					default: (
						fseek f SubChunkTextureGroup.DataSize #seek_cur
					)
				)
			)
		)
		(0x10): ( -- Scene Node
			while (not ftell f >= CalculateChunkEndOffset SubChunkMain) do
			(
				SubChunkSceneNode = ReadChunk f
				--PrintChunk SubChunkSceneNode
				
				SceneInfo_Array = #()
					
				case  (SubChunkSceneNode.DataType) of
				(
					(0x1): ( -- Scene Info
						SceneInfo = (ModelInfoStruct \
							ModelCount:(readlong f) \
							Unknown1:(readlong f) \
							Unknown2:(readlong f) \
						)
						--print SceneInfo
						append SceneInfo_Array SceneInfo
					)
					(0xE): ( --- Bones Node
						while (not ftell f >= CalculateChunkEndOffset SubChunkSceneNode) do
						(
							SubChunkBoneNode = ReadChunk f
							--PrintChunk SubChunkBoneNode
							
							case (SubChunkBoneNode.DataType) of
							(
								(0x1): ( -- Bone Data
									while (not ftell f >= CalculateChunkEndOffset SubChunkBoneNode) do
									(
										--print BoneIndexStart
										BoneCount = readlong f
										for b = 1 to BoneCount do
										(
											m11 = readfloat f; m12 = readfloat f; m13 = readfloat f
											m21 = readfloat f; m22 = readfloat f; m23 = readfloat f
											m31 = readfloat f; m32 = readfloat f; m33 = readfloat f
											m41 = readfloat f; m42 = readfloat f; m43 = readfloat f
											BoneParent = (readlong f) + BoneIndexStart
											--print ("actual bone parent "+(BoneParent - BoneIndexStart) as string)
											BoneFlag = readlong f -- Always 3? except for levels
											BoneName = ("Bone"+((b + BoneIndexStart) as string))
											--print (BoneName+" BoneFlag = "+BoneFlag as string)
											
											tfm = matrix3 [m11,m12,m13] [m21,m22,m23] [m31,m32,m33] [m41,m42,m43]
											append BoneDataArray (BoneStruct BoneParent:BoneParent BoneName:BoneName BoneTFM:tfm)
										)
									)
								)
								default: (
									fseek f (SubChunkBoneNode.DataSize) #seek_cur
								)
							)
							
							for i = 1 to BoneCount do
							(
								local extensionNode = ReadChunk f
								
								if (extensionNode.DataSize != 0) then
								(
									local hAnimNode = ReadChunk f
									local hAnimVersion = ReadLong f
									local hAnimNameId = ReadLong f
									local hAnimNumNodes = ReadLong f
									
									if (hAnimNumNodes != 0) then
									(
										local hAnimFlags = ReadLong f
										local hAnimKeyFrameSize = ReadLong f
										for j = 1 to hAnimNumNodes do
										(
											local hAnimNodeNameId = ReadLong f
											local hAnimNodeIndex = ( ReadLong f ) + 1
											local hAnimNodeFlags = ReadLong f
                                            
                                            local nodeInfo = HAnimNodeInfo()
                                            nodeInfo.NameId = hAnimNodeNameId
                                            nodeInfo.Index = hAnimNodeIndex
                                            append hAnimNodeInfoArray nodeInfo
										)
									)
                                    
                                    local curNodeInfo = undefined
                                    for nodeInfo in hAnimNodeInfoArray do
                                    (
                                        if ( nodeInfo.NameId == hAnimNameId ) then
                                        (
                                            curNodeInfo = nodeInfo
                                            exit
                                        )
                                    )
                                    
                                    if ( curNodeInfo != undefined ) then
                                        curNodeInfo.FrameIndex = i
                                    
									BoneDataArray[i].BoneName = "_" + hAnimNameId as string
								)
                                else
                                (
                                    BoneDataArray[i].BoneName = "RootNode"
                                )
							)
							
							
						
						)
					)
					(0x1A): ( -- Model Node
						while (not ftell f >= CalculateChunkEndOffset SubChunkSceneNode) do
						(
							SubChunkModelNode = ReadChunk f
							--PrintChunk SubChunkModelNode
							
							ModelInfoArray = #()
							
							ModelCounter += 1
							--print ("Model Counter = "+ModelCounter as string)
							
							case  (SubChunkModelNode.DataType) of
							(
								(0xF): ( -- Mesh Node
									while (not ftell f >= CalculateChunkEndOffset SubChunkModelNode) do
									(
										SubChunkMeshNode = ReadChunk f
										--PrintChunk SubChunkMeshNode
										
										MaterialGroupInfoArray = #()
										MaterialNameArray = #()
										MaterialNameCount = 0
										
										case  (SubChunkMeshNode.DataType) of
										(
											(0x1): ( -- Mesh Data
												while (not ftell f >= CalculateChunkEndOffset SubChunkMeshNode) do
												(
													ArrayFormat = readshort f #unsigned -- maybe 1 uint32 instead of 2 shorts
													UVCount = readshort f #unsigned
													FaceCount = readlong f
													VertCount = readlong f
													Unknown1 = readlong f -- Maybe a count, amount of buffers maybe?
													
													
													ColorFlag = bit.get ArrayFormat 4
													-- Don't really know the other flags, nor does it seem necessary to know (unless you want to export back D;)
													
													-- Array flags observed:
													-- 0x10037, 0x10077 -- UV, Faces, BBox, Vertices, Normals
													-- 0x1003F, 0x1007F -- Colors, UV, Faces, BBox, Vertices, Normals
													-- 0x200B3 -- UV, UV2, Faces, BBox, Vertices, Normals
													-- 0x200FB -- Colors, UV, UV2, Faces, BBox, Vertices, Normals
													-- 0x00033, 0x00073 -- Faces, BBox, Vertices, Normals
													-- 0x0003B, 0x0007B -- Colors, Faces, BBox, Vertices, Normals
													
													
													Color_Array = #()
													UVs_Array = #()
													Face_Array = #()
													MatID_Array = #()
													Vert_Array = #()
													Normal_Array = #()
													BoundingBox_Array = #()
													
													if (ColorFlag == true) then
													(
														for v = 1 to VertCount do 
														(
															ColorR = readbyte f #unsigned
															ColorG = readbyte f #unsigned
															ColorB = readbyte f #unsigned
															ColorA = readbyte f #unsigned
															append Color_Array [ColorR,ColorG,ColorB]
														)
													)
													for x = 1 to UVCount do
													(
														UV_Array = #()
														for v = 1 to VertCount do 
														(
															tu = (readfloat f) 
															if (gUVFlip) then tv = ((readfloat f) *-1)
															else tv = (readfloat f)
															append UV_Array [tu,tv,0]
														)
														UVs_Array[x] = (UVStruct UV_Array)
													)
													for y = 1 to FaceCount do 
													(
														f1 = readshort f + 1 -- 3ds max's indices start from 1, not 0
														f2 = readshort f + 1
														MatID = readshort f + 1
														f3 = readshort f + 1
														append Face_Array [f3,f2,f1] -- faces are stored backwards
														append MatID_Array MatID
														--print ("f1 "+f1 as string+" f2 "+f2 as string+" f3 "+f3 as string+" f4 "+f4 as string)
													)
													BoundingBoxPos = (BoundingBoxStruct \
														minX:(readfloat f) \
														minY:(readfloat f) \
														minZ:(readfloat f) \
														maxX:(readfloat f) \
														maxY:(readfloat f) \
														maxZ:(readfloat f) \
													)
													append BoundingBox_Array BoundingBoxPos
													for x = 1 to VertCount do 
													(
														vx = readfloat f
														vy = readfloat f
														vz = readfloat f
														append Vert_Array [vx,vy,vz]
													)

													for z = 1 to VertCount do 
													(
														nx = readfloat f
														ny = readfloat f
														nz = readfloat f
														append Normal_Array [nx,ny,nz]
													)
													
													append MeshDataArray (MeshData Color_Array:Color_Array UVs_Array:UVs_Array Face_Array:Face_Array MatID_Array:MatID_Array Vert_Array:Vert_Array Normal_Array:Normal_Array)
													
													fseek f (CalculateChunkEndOffset SubChunkMeshNode) #seek_set
													
												)
												MeshCounter += 1
												--print ("MeshCounter = "+MeshCounter as string)
											)
											(0x8): ( -- Material Group Node
												while (not ftell f >= CalculateChunkEndOffset SubChunkMeshNode) do
												(
													SubChunkMaterialGroupNode = ReadChunk f
													--PrintChunk SubChunkMaterialGroupNode
													MaterialNameRead = 0
													
													case  (SubChunkMaterialGroupNode.DataType) of
													(
														(0x1): ( -- Sub data
															append MaterialGroupInfoArray (ReadMatProperty SubChunkMaterialGroupNode f)
															--print MaterialGroupInfoArray
														)
														(0x7): ( -- Material Node
															while (not ftell f >= CalculateChunkEndOffset SubChunkMaterialGroupNode) do
															(
																SubChunkMaterialNode = ReadChunk f
																--PrintChunk SubChunkMaterialNode
																
																--print ("MaterialCounter = "+MaterialCounter as string)
																		
																case (SubChunkMaterialNode.DataType) of
																(
																	(0x1): ( -- Material Info
																		fseek f (SubChunkMaterialNode.DataSize) #seek_cur
																	)
																	(0x6): ( -- Material Texture Node
																		while (not ftell f >= CalculateChunkEndOffset SubChunkMaterialNode) do
																		(
																			SubChunkMaterialTextureNode = ReadChunk f
																			--PrintChunk SubChunkMaterialTextureNode
																			
																					
																			case (SubChunkMaterialTextureNode.DataType) of
																			(
																				(0x1): ( -- Material Texture Info
																					MaterialTextureInfo = readlong f
																				)
																				(0x2): ( -- Material Texture Name Info
																					MaterialNameRead += 1
																					MaterialName = readfixedstring f SubChunkMaterialTextureNode.DataSize
																					if MaterialNameRead == 1 then (
																					append MaterialNameArray MaterialName -- only store the diffuse map
																					)
																				)
																				default: (
																					fseek f (SubChunkMaterialTextureNode.DataSize) #seek_cur
																				)
																			)
																		)
																	)
																	default: (
																		fseek f (SubChunkMaterialNode.DataSize) #seek_cur
																	)
																)
															)
															MaterialCounter += 1
														)
														default: (
															fseek f (SubChunkMaterialGroupNode.DataSize) #seek_cur
														)
													)
												)
												append MaterialNamesArray MaterialNameArray
											)
											(0x3): ( -- Weight Group Node
												while (not ftell f >= CalculateChunkEndOffset SubChunkMeshNode) do
												(
													SubChunkWeightGroupNode = ReadChunk f
													--PrintChunk SubChunkWeightGroupNode
													
													case (SubChunkWeightGroupNode.DataType) of
													(
														(0x50E): ( -- Vertex IDs ???
															fseek f (SubChunkWeightGroupNode.DataSize) #seek_cur
															/*
															unk1 = readlong f
															unk2 = readlong f
															EntryCount = readlong f
															unk3 = readlong f
															unk4 = readlong f -- always null?
															*/
														)
														(0x11D): ( -- ???
															while (not ftell f >= CalculateChunkEndOffset SubChunkWeightGroupNode) do
															(
																unk1 = readlong f
																SubChunkUnknown1 = ReadChunk f
																--PrintChunk SubChunkUnknown1
																
																case (SubChunkUnknown1.DataType) of
																(
																	(0x2C): ( -- ???
																		while (not ftell f >= CalculateChunkEndOffset SubChunkUnknown1) do
																		(
																			SubChunkUnknown2 = ReadChunk f
																			--PrintChunk SubChunkUnknown2
																			
																			case (SubChunkUnknown2.DataType) of
																			(
																				(0x1): ( -- ???
																					while (not ftell f >= CalculateChunkEndOffset SubChunkUnknown2) do
																					(
																						--print "Unknown data, floats above 1??"
																						unk11 = readlong f -- always 0x1?
																						fseek f 0x18 #seek_cur -- Is this a matrix or some kind of transform?
																						IndexCount = readlong f
																						WeightCount = readlong f
																						for x = 1 to WeightCount do
																						(
																							unk1 = readbyte f
																							unk2 = readbyte f
																							unk3 = readbyte f
																							unk4 = readbyte f
																							unk5 = readfloat f
																							unk6 = readbyte f
																							unk7 = readbyte f
																							unk8 = readbyte f
																							unk9 = readbyte f
																							unk10 = readfloat f
																							--print ("Weight floats: "+ unk5 as string +" "+ unk10 as string)
																						)
																						for x = 1 to IndexCount do
																						(
																							unk11 = readshort f
																							--print ("Weight indices: "+ unk11 as string)
																						)
																					)
																				)
																				default: (
																					fseek f (SubChunkUnknown2.DataSize) #seek_cur
																				)
																			)
																		)
																	)
																	default: (
																		fseek f (SubChunkWeightsNode.DataSize) #seek_cur
																	)
																)
															)
														)
														(0x116): ( -- Weight Data
															while (not ftell f >= CalculateChunkEndOffset SubChunkWeightGroupNode) do
															(
																Weight_Array = #()
																WeightBoneArray = #()
																BonePallet_Array = #()
																UsedBoneCount = readbyte f #unsigned
																UsedBoneIndexCount = readbyte f #unsigned
																flag = readshort f -- 0x3 has extra indices after the weights
																--print ("flag = "+flag as string)
																for x = 1 to UsedBoneIndexCount do
																(
																	unk = readbyte f -- appears to be unused by the game
																)
																--Print ("Last Read @ 0x"+((bit.intAsHex(ftell f))as string))
																for x = 1 to VertCount do
																(
																	bone1 = readbyte f #unsigned
																	bone2 = readbyte f #unsigned
																	bone3 = readbyte f #unsigned
																	bone4 = readbyte f #unsigned
																	append WeightBoneArray (WeightBoneIds bone1:bone1 bone2:bone2 bone3:bone3 bone4:bone4)
																)
																--Print ("Last Read @ 0x"+((bit.intAsHex(ftell f))as string))
																for x = 1 to VertCount do
																(
																	weight1 = readfloat f
																	weight2 = readfloat f
																	weight3 = readfloat f
																	weight4 = readfloat f
																	
																	w = (weight_data boneids:#() weights:#())
																	maxweight = 0
																	if(weight1 != 0.0) then
																	   maxweight = maxweight + weight1
																	if(weight2 != 0.0) then
																	   maxweight = maxweight + weight2
																	if(weight3 != 0.0) then
																	   maxweight = maxweight + weight3
																	if(weight4 != 0.0) then
																	   maxweight = maxweight + weight4


																	if(maxweight != 0.0) then (
																		  if(weight1 != 0.0) then (
																			 w1 = weight1
																			 append w.boneids (WeightBoneArray[x].bone1 + 1)
																			 append w.weights (w1)
																		  )
																		  if(weight2 != 0.0) then (
																			 w2 = weight2
																			 append w.boneids (WeightBoneArray[x].bone2 + 1)
																			 append w.weights (w2)
																		  )
																		  if(weight3 != 0.0) then (
																			 w3 = weight3
																			 append w.boneids (WeightBoneArray[x].bone3 + 1)
																			 append w.weights (w3)
																		  )
																		  if(weight4 != 0.0) then (
																			 w4 = weight4
																			 append w.boneids (WeightBoneArray[x].bone4 + 1)
																			append w.weights (w4)
																		  )      
																	)
																	append Weight_Array w
																	--print Weight_Array
																)
																Print ("Last Read @ 0x"+((bit.intAsHex(ftell f))as string))
																--print Weight_Array
																for x = 1 to UsedBoneCount do
																(
																	fseek f 0x40 #seek_cur -- bone matrix reference
																)
																unk3 = readlong f
																unk4 = readlong f
																unk5 = readlong f
																if (flag == 0x3) then
																(
																	--Print ("Last Read @ 0x"+((bit.intAsHex(ftell f))as string))
																	for x = 1 to UsedBoneCount do
																	(
																		unk6 = readbyte f
																	)
																	--Print ("Last Read @ 0x"+((bit.intAsHex(ftell f))as string))
																	for x = 1 to UsedBoneIndexCount do
																	(
																		unk7 = readshort f
																	)
																	--Print ("Last Read @ 0x"+((bit.intAsHex(ftell f))as string))
																)
																fseek f (CalculateChunkEndOffset SubChunkWeightGroupNode) #seek_set -- to make sure it reaches the end of the chunk properly
																--Print ("Last Read @ 0x"+((bit.intAsHex(ftell f))as string))
																WeightDataInfoArray[MeshCounter] = (WeightDataInfo UsedBoneCount:UsedBoneCount UsedBoneIndexCount:UsedBoneIndexCount Weight_Array:Weight_Array)
															)
														)
														default: (
															fseek f (SubChunkWeightGroupNode.DataSize) #seek_cur
														)
													)
												)
											)
											(0x11F): ( -- User Data
												fseek f (SubChunkMeshNode.DataSize) #seek_cur
											)
											default: (
												fseek f (SubChunkMeshNode.DataSize) #seek_cur
											)
										)
									)
								)
								default: (
									fseek f (SubChunkModelNode.DataSize) #seek_cur
								)
							)
						)
					)
					(0x14): ( -- Draw Node
						while (not ftell f >= CalculateChunkEndOffset SubChunkSceneNode) do
						(
							SubChunkSkinNode = ReadChunk f
							--PrintChunk SubChunkSkinNode
							
							RootBoneArray = #()
							
							case (SubChunkSkinNode.DataType) of
							(
								(0x1): ( -- Draw Data
									while (not ftell f >= CalculateChunkEndOffset SubChunkSkinNode) do
									(
										RootBoneID = (readlong f + 1) + BoneIndexStart  
										ModelID = (readlong f + 1) + ModelIndexStart
										Flag1 = readlong f + 1 -- Always 0x6
										Flag2 = readlong f + 1 -- Always 0x1
										append RootBoneArray (SkinDataStruct RootBoneID:RootBoneID ModelID:ModelID Flag1:Flag1 Flag2:Flag2)
									)
									append RootBonesArray (RootBoneStruct RootBoneArray:RootBoneArray)
								)
								default: (
									fseek f (SubChunkSkinNode.DataSize) #seek_cur
								)
							)
						)
					)
					default: (
						fseek f (SubChunkSceneNode.DataSize) #seek_cur
					)
				)
			)
		)
		(0xFFFFFFFFF0F000E0): ( -- Particle Group Node
			fseek f (SubChunkMain.DataSize + 0x2) #seek_cur
		)
		default: (
			--print "SKIP"
			fseek f (SubChunkMain.DataSize) #seek_cur
		)
	)
	ModelIndexStart += MeshCount
	BoneIndexStart += BoneCount
	--print ("modelidxstart = "+ModelIndexStart as string)
)

for b = 1 to BoneDataArray.count do
(		
	if (getNodeByName BoneDataArray[b].BoneName) != undefined do (append BoneArray (getNodeByName BoneDataArray[b].BoneName))
	if ((BoneDataArray[b].BoneParent != -1) AND (BoneDataArray[b].BoneParent < b)) then ( BoneDataArray[b].BoneTFM = BoneDataArray[b].BoneTFM * BoneArray[BoneDataArray[b].BoneParent+1].objecttransform ) 
	else if ((b > 1) AND (BoneDataArray[b].BoneParent > -1)) do (print ("Bone named " + BoneDataArray[b].BoneName as string + " has a parent bone that doesn't exist yet!"))
	
	newBone = bonesys.createbone	\
					  BoneDataArray[b].BoneTFM.row4	\
					  (BoneDataArray[b].BoneTFM.row4 + 0.01 * (normalize BoneDataArray[b].BoneTFM.row1)) \
					  (normalize BoneDataArray[b].BoneTFM.row3)
				newBone.name = BoneDataArray[b].BoneName
				newBone.width  = 0.01
				newBone.height = 0.01
				newBone.transform = BoneDataArray[b].BoneTFM
				newBone.setBoneEnable false 0
				newBone.wirecolor = yellow
				newbone.showlinks = true
				newBone.pos.controller      = TCB_position ()
				newBone.rotation.controller = TCB_rotation ()

	if (BoneDataArray[b].BoneParent != -1) then newBone.parent = BoneArray[(BoneDataArray[b].BoneParent+1)]
                          
    local curNodeInfo = undefined
    for nodeInfo in hAnimNodeInfoArray do
    (
        if ( nodeInfo.FrameIndex == b ) then
        (
            curNodeInfo = nodeInfo
            exit
        )
    )
    
    if ( curNodeInfo != undefined ) then
    (
        SetUserProp newBone "NameId" ( curNodeInfo.NameId as string )
        SetUserProp newBone "Index" ( curNodeInfo.Index as string )
    )
    
    SetUserProp newBone "FrameIndex" ( b as string )
                          
	append BoneArray newBone
)

for x = 1 to RootBonesArray.count do
(
	ModelID = (RootBonesArray[x].RootBoneArray[1].ModelID)
	msh = mesh Vertices:MeshDataArray[ModelID].Vert_Array Faces:MeshDataArray[ModelID].Face_Array
	msh.numTVerts = MeshDataArray[ModelID].Vert_Array.count
	msh.name = filename+"_"+(ModelID) as string
	buildTVFaces msh
	for a = 1 to MeshDataArray[ModelID].UVs_array.count do (for j = 1 to MeshDataArray[ModelID].UVs_array[1].UV_Array.count do setTVert msh j MeshDataArray[ModelID].Uvs_Array[1].UV_array[j])
	for j = 1 to MeshDataArray[ModelID].Face_array.count do setTVFace msh j MeshDataArray[ModelID].Face_array[j]
	for j = 1 to msh.numfaces do setFaceSmoothGroup msh j 1
	for j = 1 to msh.numfaces do setFaceMatID msh j (MeshDataArray[ModelID].MatID_Array[j] as integer)
	
	if (gImportNormals) then
	(
		max modify mode
		select msh
		addmodifier msh (Edit_Normals ()) ui:off
		msh.Edit_Normals.MakeExplicit selection:#{1..MeshDataArray[ModelID].Normal_array.count}
		EN_convertVS = msh.Edit_Normals.ConvertVertexSelection
		EN_setNormal = msh.Edit_Normals.SetNormal
		normID = #{}

		for v = 1 to MeshDataArray[ModelID].Normal_array.count do(
			free normID
			EN_convertVS #{v} &normID
			for id in normID do EN_setNormal id MeshDataArray[ModelID].Normal_array[v]
		)
		maxOps.CollapseNodeTo msh 1 true -- Collapse the Edit Normals modifier
	)
	
	setNumCPVVerts msh msh.numTVerts
	defaultVCFaces msh
	for j = 1 to MeshDataArray[ModelID].Color_Array.count do setvertcolor msh j MeshDataArray[ModelID].Color_array[j]
	
	msh.material = multimaterial numsubs:MaterialNamesArray[ModelID].count
	--print ("matname "+MaterialNamesArray as string)
	for j = 1 to msh.material.count do 
	(
		if (doesFileExist (filePath + filename + "_tex\\" + MaterialNamesArray[ModelID][j] as string+".png")) then (
			tm = Bitmaptexture filename:(filePath + filename + "_tex\\" + MaterialNamesArray[ModelID][j] as string+".png")
			tm.alphasource = 0
			tm.monoOutput = 1
			msh.material.materialList[j].diffuseMap = tm
		)
		else msh.material.materialList[j].Diffuse = random white black
		msh.material.materialList[j].name = (MaterialNamesArray[ModelID][j] as string)
		msh.material.materialList[j].showInViewport = true
	)
	
	rootBone = getNodeByName BoneDataArray[(RootBonesArray[x].RootBoneArray[1].RootBoneID)].BoneName
	msh.transform = rootBone.transform
	msh.backfaceCull = true
	resetxform msh
	maxOps.CollapseNode msh off
	if (WeightDataInfoArray[x] == undefined AND gImportWeights == true) then
	(
		max modify mode
		select msh
		skinMod = skin ()
		addModifier msh skinMod
		skinOps.addBone skinMod rootBone 1 
		modPanel.setCurrentObject skinMod
	)
)

if (gImportWeights) then
(
	for x = 1 to MeshCounter do
	(
		if (not WeightDataInfoArray[x] == undefined) then
		(
			max modify mode
			msh = getNodeByName (filename+"_"+x as string) exact:true
			select msh
			skinMod = skin ()
			addModifier msh skinMod
            
			for i = 1 to WeightDataInfoArray[x].UsedBoneCount do
			( 
                local curNodeInfo = undefined
                for nodeInfo in hAnimNodeInfoArray do
                (
                    if ( nodeInfo.Index == i ) then
                    (
                        curNodeInfo = nodeInfo
                        exit
                    )
                )
                
                local maxBone = GetNodeByName ( "_" + curNodeInfo.NameId as string ) exact:true
                
				--print maxBone
				if i != WeightDataInfoArray[x].UsedBoneCount then 
					skinOps.addBone skinMod maxbone 0
				else
					skinOps.addBone skinMod maxbone 1
			)
			modPanel.setCurrentObject skinMod
			
			for i = 1 to (getnumverts msh) do (
			   w = WeightDataInfoArray[x].Weight_Array[i]
			   bi = #() --bone index array
			   wv = #() --weight value array
			   
			   for j = 1 to w.boneids.count do
			   (
				  boneid = w.boneids[j]
				  weight = w.weights[j]
				  append bi boneid
				  append wv weight
			   )   
			   skinOps.ReplaceVertexWeights skinMod i bi wv
			)
		)
	)
)

GC()
FClose f
EnableSceneRedraw()

Print ("Done! ("+((((timestamp())-st)*0.001)as string)+" Seconds)")
